# ----------------------------------------------------------------------
# METADATA EXPORT INTO GEONETWORK WITH GEOMETA AND GEONAPI LIBRARIES
# author : Juliette Fabre / OSU OREME
# creation : 04/18
# last update : 19/19
# -----------------------------------------------------------------------

# The script extracts all metadata from the appropriates views, formats them into ISO19115 with geometa library, and inserts them into the OSU OREME Geonetwork (GN) catalog  with geonapi library.
# The script is based on the geometa ISOMetadata example (https://github.com/eblondel/geometa/blob/master/inst/extdata/examples/metadata.R) and geonapi example (see geonapi documentation: https://github.com/eblondel/geonapi/wiki)

# Emmanuel Blondel. (2019, May 26). geometa: Tools for Reading and Writing ISO/OGC Geographic Metadata in R (Version 0.5). Zenodo. http://doi.org/10.5281/zenodo.3229292
# Blondel, Emmanuel. (2019, May 26). geonapi: R Interface to GeoNetwork API (Version 0.2-1). Zenodo. http://doi.org/10.5281/zenodo.3237237

# SCRIPT PROGRESS
# ---------------

# - extract metadata from all views <my_schema>.<my_view> that contain at least the following fields (see the view extraction example metadata_view.csv):
#   - resource_identifier (home-made identifier)
#   - title
#   - abstract
#   - creation_date (YYYY-MM-DD)
#   - start_date (YYYY-MM-DD)
#   - end_date (YYYY-MM-DD)
#   - geom = geometry WKT in WGS84 (eg in postgis: st_astext(st_envelope(st_convexhull(st_collect(geom)))) )
#   - west, east, south, north: min longitude, max longitude, min latitude, max latitude
#   - resource_type (eg: "Dataset")
#   - use_condition = list of use conditions separated with ---
#   - resource_language (eg:  "eng")
#   - resource_format (eg:  "CSV")
#   - inspire_themes = list of <keyword>---<uri> separated with ,
#   - gemet_keywords = list of <keyword>---<uri> separated with ,
#   - agrovoc_keywords = list of <keyword>---<uri> separated with ,
#   - other_keywords = list of keywords separated with ,
#   - md_contact (responsible) = list of <name, first name>---<organisation>---<email> separated with ;
#   - resource_contacts (= optional more precise contacts) = list of <name, first name>---<organisation>---<email> separated with ;
#   - topic_categories  = list of topics separated with , (= environment, geoscientificInformation)
#   - update_frequency (eg: "continual" / "notPlanned")
#   - status (eg: onGoing / completed)
#   - online_resource = list of <url>---<description> separated with ,
#   - wms_resource = list of <wms_url>---<layer_name>---<description> separated with ,
#   - lineage
#   - parent_identifier
#   - reference_system (EPSG code, eg: 4326)
#   - thumbnail_url (thumbnail url)

# NB: You can use either the geom field or the west, east, south, north fields

# - add a unique universal identifier to each metadata: if the metadata already exists in GN, extract the corresponding uuid from the gn_record table ; if not, create a new uuid
# - format metadata into XML files with Geometa library and export then into GN with Geonapi library
# - If the metadata was inserted, register the metadata and its uuid into the gn_record table

# Get arguments
args <- commandArgs()
cleanargs <- sub("-.*=", "", args)
dir <- cleanargs[regexpr("-d=", args)>0]

# Load function files (available on https://data.oreme.org/tools/R)
script_path <- dirname(sub("--file=", "", args[4]))
source(paste0(script_path, "/oreme_db_v1.4.R"))
source(paste0(script_path, "/oreme_tools_v1.2.R"))

# LOAD LIBRARIES
library(XML)
library(geometa)
library(geonapi)
library(uuid)
library(rgeos)

# CONNECT TO OREME DATABASE 
# Homemade function (see oreme_db_v1.4.R)
connexion <- connect_db(my_pg_dbname, my_user, my_password, my_host) 
con <- connexion$con

# GN CONFIGURATION
gn_url <- "http://162.38.140.226:8080/geonetwork"
GN <- GNManager$new(url = gn_url, version = "3.0.3.0", user = "my_login", pwd = "my_pwd", logger = "INFO")

# Several views (in different schemas) are parsed to generate the Oreme catalogue. In this example only one view will be parsed.
my_schema <- 'public'
my_view <- 'metadata_view'

# GET GEONETWORK RECORDS
# The gn_record table records all GN metadata, with their GN identifier and their internal identifier. It contains the fields:
# - oreme_identifier: homemade identifier, as created in all metadata_view (one for each metadata record)
# - md_identifier: GN identifier (uuid)

# homemade function to extract all records of the table (see oreme_db_v1.4.R)
gn_records <- get_db_values('*', 'gn_record', con = con)

cat(paste0('<b>--- SCHEMA ', my_schema, ', VUE ', my_view, '</b>\n'))

# SELECT METADATA
# homemade function to extract all records of the table "my_view" from the "my_schema" schema (see oreme_db_v1.4.R)
metadata <- get_db_values('*', my_view, my_schema, con)

# For temporalExtent XSD validation in GN
ISOMetadataNamespace[["GML"]]$uri <- "http://www.opengis.net/gml"

for(i in 1:nrow(metadata))
{
  cat(paste0('<b>--------- ', metadata$resource_identifier[i], '</b>\n'))
  md <- ISOMetadata$new()
  
  # ------------------------------------
  # METADATA SECTION
  # ------------------------------------
  
  # MD identifier
  resource_identifier <- metadata$resource_identifier[i]
  # Check if the metadata have already been inserted into GN or not
  if(nrow(gn_records) && resource_identifier %in% gn_records$oreme_identifier) md_identifier <- gn_records$md_identifier[gn_records$oreme_identifier == resource_identifier] else md_identifier <- UUIDgenerate()
  md$setFileIdentifier(md_identifier)
  
  # Parent Uuid
  if(!is.na(metadata$parent_identifier[i])) md$setParentIdentifier(metadata$parent_identifier[i])
  
  md$setDataSetURI(resource_identifier)
  
  md$setLanguage("eng")
  md$setCharacterSet("utf8")
  mdDate <- Sys.Date()
  mdDate <- ISOdate(format(mdDate, "%Y"), format(mdDate, "%m"), format(mdDate, "%d"))
  md$setDateStamp(mdDate)
  md$setMetadataStandardName("ISO 19115:2003/19139")
  md$setMetadataStandardVersion("1.0")
  
  # HierarchyLevel
  md$setHierarchyLevel(tolower(metadata$resource_type[i]))
  
  # Metadata contacts
  contact_list <- strsplit(metadata$md_contact[i], ';')[[1]]
  for(c in 1:length(contact_list))
  {
    # trim => homemade function (see oreme_tools_v1.2.R)
    person <- trim(contact_list[c])
    name <- (strsplit(person, '---'))[[1]][1]
    org <- (strsplit(person, '---'))[[1]][2]
    email <- (strsplit(person, '---'))[[1]][3]
    rp <- ISOResponsibleParty$new()
    rp$setIndividualName(name)
    rp$setOrganisationName(org)
    rp$setRole("pointOfContact")
    contact <- ISOContact$new()
    address <- ISOAddress$new()
    address$setEmail(email)
    contact$setAddress(address)
    rp$setContactInfo(contact)
    md$addContact(rp)
  }
  
  # -------------------------------------
  # VECTOR SPATIAL REPRESENTATION SECTION
  # -------------------------------------
  
  vsr <- ISOVectorSpatialRepresentation$new()
  vsr$setTopologyLevel("geometryOnly")
  geomObject <- ISOGeometricObjects$new()
  geomObject$setGeometricObjectType("surface")
  geomObject$setGeometricObjectCount(5L)
  vsr$setGeometricObjects(geomObject)
  md$addSpatialRepresentationInfo(vsr)
  
  # ------------------------------------
  # REFERENCE SYSTEM SECTION
  # ------------------------------------
  
  RS <- ISOReferenceSystem$new()
  RSId <- ISOReferenceIdentifier$new(code = metadata$reference_system[i], codeSpace = "EPSG")
  RS$setReferenceSystemIdentifier(RSId)
  md$setReferenceSystemInfo(RS)
  
  # ------------------------------------
  # IDENTIFICATION SECTION
  # ------------------------------------
  
  IDENT <- ISODataIdentification$new()
  
  # Abstract
  IDENT$setAbstract(metadata$abstract[i])
  
  IDENT$setLanguage("eng")
  IDENT$setCharacterSet("utf8")
  
  # Topic categories
  # trim => homemade function (see oreme_tools_v1.2.R)
  topic_list <- trim(unlist(strsplit(metadata$topic_categories[i], ',')))
  for(t in topic_list) IDENT$addTopicCategory(t)
  
  # Status
  IDENT$addStatus(metadata$status[i])
  
  # Resource contacts
  contact_list <- strsplit(metadata$resource_contacts[i], ';')[[1]]
  for(c in 1:length(contact_list))
  {
    # trim => homemade function (see oreme_tools_v1.2.R)
    person <- trim(contact_list[c])
    name <- (strsplit(person, '---'))[[1]][1]
    org <- (strsplit(person, '---'))[[1]][2]
    email <- (strsplit(person, '---'))[[1]][3]
    rp <- ISOResponsibleParty$new()
    rp$setIndividualName(name)
    rp$setOrganisationName(org)
    rp$setRole("pointOfContact")
    contact <- ISOContact$new()
    address <- ISOAddress$new()
    address$setEmail(email)
    contact$setAddress(address)
    rp$setContactInfo(contact)
    IDENT$addPointOfContact(rp)
  }
  
  # Title
  ct <- ISOCitation$new()
  ct$setTitle(metadata$title[i])
  iso_date <- ISODate$new()
  iso_date$setDate(as.Date(mdDate))
  iso_date$setDateType("revision")
  ct$addDate(iso_date)
  if(!is.na(metadata$creation_date[i]))
  {
    iso_date <- ISODate$new()
    iso_date$setDate(as.Date(metadata$creation_date[i]))
    iso_date$setDateType("creation")
    ct$addDate(iso_date)
  }
  ct$setIdentifier(ISOMetaIdentifier$new(code = md_identifier))
  IDENT$setCitation(ct)
  
  # Thumbnail
  if(!is.na(metadata$thumbnail_url[i]))
  {
    go <- ISOBrowseGraphic$new(
      fileName = metadata$thumbnail_url[i],
      fileDescription = "Thumbnail",
      fileType = paste0("image/", get_file_ext(get_file_without_path(metadata$thumbnail_url[i])))
    )
    IDENT$addGraphicOverview(go)
  }
  
  # Maintenance information
  mi <- ISOMaintenanceInformation$new()
  mi$setMaintenanceFrequency(metadata$update_frequency[i])
  IDENT$setResourceMaintenance(mi)

  # Access legal constraints
  # for INSPIRE controlled terms on access legal constraints, please browse the INSPIRE registry:
  # http://inspire.ec.europa.eu/metadata-codelist/LimitationsOnPublicAccess/
  lc <- ISOLegalConstraints$new()
  lc$addAccessConstraint("copyright")
  lc$addAccessConstraint("license")
  lc$addAccessConstraint("otherRestrictions")
  lc$addOtherConstraint(ISOAnchor$new(
    href = "http://inspire.ec.europa.eu/metadata-codelist/LimitationsOnPublicAccess/noLimitations",
    name = "no limitations to public access"
  ))
  IDENT$addResourceConstraints(lc)
  
  # Use legal constraints
  # for INSPIRE controlled terms on use legal constraints, please browse the INSPIRE registry:
  # http://inspire.ec.europa.eu/metadata-codelist/ConditionsApplyingToAccessAndUse
  lc2 <- ISOLegalConstraints$new()
  # trim => homemade function (see oreme_tools_v1.2.R)
  use_conditions <- trim(unlist(strsplit(metadata$use_condition[i], '---')))
  if(!is.na(metadata$use_condition[i])) for(use in use_conditions) lc2$addUseLimitation(use)
  lc2$addUseLimitation('This work is licensed under a Creative Commons Attribution 4.0 License (CC BY 4.0, https://creativecommons.org/licenses/by/4.0).')
  lc2$addAccessConstraint("otherRestrictions")
  lc2$addOtherConstraint(ISOAnchor$new(
    href = "http://inspire.ec.europa.eu/metadata-codelist/ConditionsApplyingToAccessAndUse/noConditionsApply",
    name = "No conditions apply to access and use."
  ))
  IDENT$addResourceConstraints(lc2)
  
  # Spatial extent using the geom field
  if(!is.na(metadata$geom[i]))
  {
    extent <- ISOExtent$new()
    spatial_extent <- readWKT(metadata$geom[i])
    xmin <- spatial_extent@bbox[1,1]
    xmax <- spatial_extent@bbox[1,2]
    ymin <- spatial_extent@bbox[2,1]
    ymax <- spatial_extent@bbox[2,2]
    if(xmin == xmax)
    {
      xmin <- xmin - 0.001
      xmax <- xmax + 0.001
    }
    if(ymin == ymax)
    {
      ymin <- ymin - 0.001
      ymax <- ymax + 0.001
    }
    spatialExtent <- ISOGeographicBoundingBox$new(minx = xmin, miny=ymin, maxx=xmax, maxy=ymax)
    extent$setGeographicElement(spatialExtent)
    IDENT$addExtent(extent)
  }
  # Spatial extent using the west, east, south, north fields
  # if(!is.na(metadata$west[i]) && !is.na(metadata$east[i]) && !is.na(metadata$south[i]) && !is.na(metadata$north[i]))
  # {
  #   extent <- ISOExtent$new()
  #   spatialExtent <- ISOGeographicBoundingBox$new(minx = metadata$west[i], miny=metadata$south[i], maxx=metadata$east[i], maxy=metadata$north[i])
  #   extent$setGeographicElement(spatialExtent)
  #   IDENT$addExtent(extent)
  # }
  
  # Temporal extent
  if(!is.na(metadata$start_date[i]))
  {
    extent <- ISOExtent$new()
    time <- ISOTemporalExtent$new()
    start_date <- as.POSIXct(metadata$start_date[i])
    if(!is.na(metadata$end_date[i])) end_date <- as.POSIXct(metadata$end_date[i]) else end_date <- as.POSIXct(Sys.Date())
    temporalExtent <- GMLTimePeriod$new(beginPosition = start_date, endPosition = end_date)
    time$setTimePeriod(temporalExtent)
    extent$setTemporalElement(time)
    IDENT$addExtent(extent)
  }
  
  # Keywords
  # trim => homemade function (see oreme_tools_v1.2.R)
  inspire_themes <- trim(unlist(strsplit(metadata$inspire_themes[i], ',')))
  gemet_keywords <- trim(unlist(strsplit(metadata$gemet_keywords[i], ',')))
  agrovoc_keywords <- trim(unlist(strsplit(metadata$agrovoc_keywords[i], ',')))
  other_keywords <- trim(unlist(strsplit(metadata$other_keywords[i], ',')))
  
  if(!all(is.na(inspire_themes)))
  {
    kwd1 <- ISOKeywords$new()
    for(k in inspire_themes)
    {
      anc1 <- ISOAnchor$new(name = strsplit(k, '---')[[1]][1], href = strsplit(k, '---')[[1]][2])
      kwd1$addKeyword(anc1)
    }
    kwd1$setKeywordType("theme")
    th <- ISOCitation$new()
    th_ref <- ISOAnchor$new(name = "GEMET - INSPIRE themes, version 1.0", href="http://www.eionet.europa.eu/gemet/inspire_themes")
    th$setTitle(th_ref)
    inspire_date <- ISODate$new()
    inspire_date$setDate(ISOdate(2008,06,01))
    inspire_date$setDateType("publication")
    th$addDate(inspire_date)
    kwd1$setThesaurusName(th)
    IDENT$addKeywords(kwd1)
  }
  if(!all(is.na(gemet_keywords)))
  {
    kwd2 <- ISOKeywords$new()
    for(k in gemet_keywords)
    {
      anc1 <- ISOAnchor$new(name = strsplit(k, '---')[[1]][1], href = strsplit(k, '---')[[1]][2])
      kwd2$addKeyword(anc1)
    }
    kwd2$setKeywordType("theme")
    th <- ISOCitation$new()
    th$setTitle("GEMET")
    gemet_date <- ISODate$new()
    gemet_date$setDate(ISOdate(2010,01,13))
    gemet_date$setDateType("publication")
    th$addDate(gemet_date)
    kwd2$setThesaurusName(th)
    IDENT$addKeywords(kwd2)
  }
  if(!all(is.na(agrovoc_keywords)))
  {
    kwd3 <- ISOKeywords$new()
    for(k in agrovoc_keywords)
    {
      anc1 <- ISOAnchor$new(name = strsplit(k, '---')[[1]][1], href = strsplit(k, '---')[[1]][2])
      kwd3$addKeyword(anc1)
    }
    kwd3$setKeywordType("theme")
    th <- ISOCitation$new()
    th$setTitle("AGROVOC")
    agrovoc_date <- ISODate$new()
    agrovoc_date$setDate(ISOdate(2011,02,23))
    agrovoc_date$setDateType("publication")
    th$addDate(agrovoc_date)
    kwd3$setThesaurusName(th)
    IDENT$addKeywords(kwd3)
  }
  if(!all(is.na(other_keywords)))
  {
    kwd4 <- ISOKeywords$new()
    for(k in other_keywords) kwd4$addKeyword(k)
    kwd4$setKeywordType("theme")
    IDENT$addKeywords(kwd4)
  }
  
  # Spatial representation type
  IDENT$addSpatialRepresentationType("vector")
  
  md$setIdentificationInfo(IDENT)
  
  # ------------------------------------
  # DISTRIBUTION SECTION
  # ------------------------------------
  
  # Online resources
  distrib <- ISODistribution$new()
  dto <- ISODigitalTransferOptions$new()
  online_resource_list <- strsplit(metadata$online_resource[i], ',')[[1]]
  if(length(online_resource_list)) for(o in 1:length(online_resource_list))
  {
    # trim => homemade function (see oreme_tools_v1.2.R)
    link <- trim(online_resource_list[o])
    url <- strsplit(link, '---')[[1]][1]
    name <- strsplit(link, '---')[[1]][2]
    if(name != 'Associated DOI')
    {
      newURL <- ISOOnlineResource$new()
      newURL$setLinkage(url)
      newURL$setName(name)
      newURL$setDescription(name)
      newURL$setProtocol("WWW:LINK-1.0-http--link")
      dto$addOnlineResource(newURL)
    }
  }
  if(!is.na(metadata$wms_resource[i]))
  {
    wms_resource_list <- strsplit(metadata$wms_resource[i], ',')[[1]]
    for(wms in 1:length(wms_resource_list))
    {
      # trim => homemade function (see oreme_tools_v1.2.R)
      link <- trim(wms_resource_list[wms])
      url <- strsplit(link, '---')[[1]][1]
      name <- strsplit(link, '---')[[1]][2]
      desc <- strsplit(link, '---')[[1]][3]
      newURL <- ISOOnlineResource$new()
      newURL$setLinkage(url)
      newURL$setName(name)
      newURL$setDescription(desc)
      newURL$setProtocol("OGC:WMS")
      dto$addOnlineResource(newURL)
    }
  }
  distrib$setDigitalTransferOptions(dto)
  
  # Format
  format <- ISOFormat$new()
  format$setName(metadata$resource_format[i])
  format$setVersion(metadata$version_format[i])
  distrib$addFormat(format)
  
  md$setDistributionInfo(distrib)
  
  # ------------------------------------
  # DATA QUALITY
  # ------------------------------------
  
  dq <- ISODataQuality$new()
  scope <- ISOScope$new()
  scope$setLevel("dataset")
  dq$setScope(scope)
  
  # Lineage
  lineage <- ISOLineage$new()
  lineage$setStatement(metadata$lineage[i])
  dq$setLineage(lineage)
  
  # INSPIRE - interoperability of spatial data sets and services
  dc_inspire1 <- ISODomainConsistency$new()
  cr_inspire1 <- ISOConformanceResult$new()
  cr_inspire_spec1 <- ISOCitation$new()
  cr_inspire_spec1$setTitle("Commission Regulation (EU) No 1089/2010 of 23 November 2010 implementing Directive 2007/2/EC of the European Parliament and of the Council as regards interoperability of spatial data sets and services")
  cr_inspire1$setExplanation("See the referenced specification")
  cr_inspire_date1 <- ISODate$new()
  cr_inspire_date1$setDate(as.Date(ISOdate(2010,12,08)))
  cr_inspire_date1$setDateType("publication")
  cr_inspire_spec1$addDate(cr_inspire_date1)
  cr_inspire1$setSpecification(cr_inspire_spec1)
  cr_inspire1$setPass(TRUE)
  dc_inspire1$addResult(cr_inspire1)
  dq$addReport(dc_inspire1)
  
  # INSPIRE - metadata
  dc_inspire2 <- ISODomainConsistency$new()
  cr_inspire2 <- ISOConformanceResult$new()
  cr_inspire_spec2 <- ISOCitation$new()
  cr_inspire_spec2$setTitle("COMMISSION REGULATION (EC) No 1205/2008 of 3 December 2008 implementing Directive 2007/2/EC of the European Parliament and of the Council as regards metadata")
  cr_inspire2$setExplanation("See the referenced specification")
  cr_inspire_date2 <- ISODate$new()
  cr_inspire_date2$setDate(as.Date(ISOdate(2008,12,3)))
  cr_inspire_date2$setDateType("publication")
  cr_inspire_spec2$addDate(cr_inspire_date2)
  cr_inspire2$setSpecification(cr_inspire_spec2)
  cr_inspire2$setPass(TRUE)
  dc_inspire2$addResult(cr_inspire2)
  dq$addReport(dc_inspire2)
  
  md$addDataQualityInfo(dq)
  
  # ------------------------------------
  # EXPORT METADATA IN XML AND GN
  # ------------------------------------
  
  file_name <- paste0(dir, "/", resource_identifier, ".xml")
  cat(saveXML(md$encode(inspire=TRUE), file_name, encoding = 'UTF-8', indent = T))
  
  # Insert or update the metadata into a geonetwork
  # An insert has to be done in 2 operations (the insert itself, and the privilege setting to "publish" it either to a restrained group or to public)
  # An update has to be done based on the internal Geonetwork id (that can be queried as well)
  privileges <- c("view", "dynamic", "featured")
  
  metaId <- GN$get(md_identifier, by = "uuid", output = "id")
  if(is.null(metaId))
  {
    # insert metadata (once inserted only visible to the publisher)
    created = GN$insertMetadata(xml = md$encode(inspire=TRUE), group = "1", category = "datasets")
    # config privileges
    config <- GNPrivConfiguration$new()
    config$setPrivileges("all", privileges)
    GN$setPrivConfiguration(id = created, config = config)
    # insert record into gn_record table
    req <- paste0("INSERT INTO public.gn_record(oreme_identifier, md_identifier) VALUES('", resource_identifier, "','", md_identifier, "')")
    res <- dbSendQuery(con, req)
  } else
  {
    # update a metadata
    updated = GN$updateMetadata(id = metaId, xml = md$encode(inspire=TRUE))
    # config privileges
    config <- GNPrivConfiguration$new()
    config$setPrivileges("all", privileges)
    GN$setPrivConfiguration(id = metaId, config = config)
  }
}